################################################################
#                  'OgCode' pathing program.                   #
#        Perpetually providing precise printer pathing.        #
# Developed by Sam Barrett, samuel.alexander.barrett@gmail.com #
################################################################

import numpy as np
import math

############################ Build parameters ############################
#Proposed standard FDM tensile settings are:
#  layerHeight=0.4;layerWidth=0.4;bulkDiameter based on average of 20 diameter measurements;
#  Eactual based on commanding 'extrude 100mm' and measuring filament feed-through.

layerHeight=0.4 #[mm]
fibreWidth=0.4 #[mm] Note: *Diameter*
fibreArea=(fibreWidth/2.0)*(layerHeight/2.0)*math.pi #Area of fibre, assuming elliptical x-section.

bulkDiameter=1.79 #[mm] Note: *Diameter* of the bulk filament.
bulkArea=math.pi*(bulkDiameter/2.0)**2 #Area of bulk filament x-section, assuming circular x-section.
dEpermm=fibreArea/bulkArea

#Account for incorrect calibration in extruder steps:
Einput=100 #Input extrusion value [mm]
Eactual=126.5 #Actual extruded value [mm]
Emod=Einput/Eactual #Modifier to tune to correct 'extrusion per mm' value
dEpermm=dEpermm*Emod

#Account for desired gap width - see 'slicing parameters' in report.
airGapHorizontal=-0.04 #Desired gap width modifier [mm], currently one-tenth of fibre diameter
airGapVertical=-0.04 #Desired gap width modifier [mm], currently one-tenth of fibre diameter
fibreWidth=fibreWidth+airGapHorizontal #fibreWidth now accounts for desired airGap
layerHeight=layerHeight+airGapVertical #layerHeight now accounts for desired airGap

########################## End Build Parameters ##########################







###################### START OF OGCODE DEFINITIONS #######################
#Note: These definitions control the pathing parameters of the printer.
#They should not need to be altered to simply print a specimen.
#Specimens are defined at the bottom of the code.


gcodeFilename = 'OGCodeOutput.gcode'
gcode=open(gcodeFilename,'w')

#Define initial coordinates on global level. Sequence is G X Y Z F E.
G='G1'
X=0.0
Y=0.0
Z=0.0
F=0.0
E=0.0

def startGcode():
    #Comment on settings:
    gcode.write(';Print Settings: \n')
    gcode.write(';layerHeight  ')
    gcode.write(str(layerHeight))
    gcode.write(',\n ;fibreWidth ')
    gcode.write(str(fibreWidth))
    gcode.write(',\n ;fibreArea ')
    gcode.write(str(fibreArea))
    gcode.write(',\n ;bulkDiameter ')
    gcode.write(str(bulkDiameter))
    gcode.write(',\n ;bulkArea ')
    gcode.write(str(bulkArea))
    gcode.write(',\n ;dEpermm ')
    gcode.write(str(dEpermm))
    gcode.write('\n')
    
    #Set units to [mm]:
    gcode.write('G21\n')

    #Set all coordinates to absolute positioning:
    gcode.write('G90\n')
    gcode.write('G90 X0 Y0 Z0\n')

    #Set E datum to 0:
    gcode.write('G92 E0\n')

    #Home X,Y axes:
    gcode.write('G28 X0 Y0 \n')

    #Fast move to center:
    gcode.write('G1 X100 Y100 F4000\n')

    #Home Z axis:
    gcode.write('G28 Z0\n')

    #Turn fan off:
    gcode.write('M107\n')

    #Set extruder to absolute positioning:
    gcode.write('M82\n')

def sendCommand(C,X,Y,Z,F,E):
    gcode.write(C)
    gcode.write(' ')
    gcode.write('X')
    gcode.write(str(X))
    gcode.write(' ')
    gcode.write('Y')
    gcode.write(str(Y))
    gcode.write(' ')
    gcode.write('Z')
    gcode.write(str(Z))
    gcode.write(' ')
    gcode.write('F')
    gcode.write(str(F))
    gcode.write(' ')
    gcode.write('E')
    gcode.write(str(E))
    gcode.write('\n')

def cylinder(pathing,xc,yc,zc,innerRadius,outerRadius,height):
    global E
    if innerRadius>outerRadius:
        print('Error: innerRadius bigger than outerRadius.')
        return
    gcode.write('\n')
    gcode.write(';Cylinder: Pathing is ')
    gcode.write(pathing)
    gcode.write(',xc ')
    gcode.write(str(xc))
    gcode.write(',yc ')
    gcode.write(str(yc))
    gcode.write(',zc ')
    gcode.write(str(zc))
    gcode.write('\n')
    gcode.write(';innerRadius ')
    gcode.write(str(innerRadius))
    gcode.write(',outerRadius ')
    gcode.write(str(outerRadius))
    gcode.write(',height ')
    gcode.write(str(height))
    gcode.write('\n')
    sliceCount=math.ceil(height/layerHeight)
    
    if pathing=='concentric':
        C='G1' #'Controlled Move' extrude-move G-code
        F=2700.000
        radialRoads=math.ceil((outerRadius-innerRadius)/fibreWidth)
        #Retract extruder:
        gcode.write('G1 F9000.000 E-2.500\n')
        gcode.write('G1 F9000.000 E0.0\n')
        for sliceIterate in np.arange(1,sliceCount+1):
            Z=zc+sliceIterate*layerHeight
            gcode.write('G92 E0\n')
            E=0.0
            for radialIterate in np.arange(radialRoads):
                pathRadius=innerRadius+radialIterate*fibreWidth
                stepsInCircle=60
                if pathRadius<2.4:
                    stepsInCircle=30
                for theta in np.linspace(0,2*math.pi,num=stepsInCircle): #Circle is divided into straight-line segments.
                    if pathRadius==0.0:
                        break
                    if theta==0.0:
                        dE=0.0
                        F=4500.000
                    else:
                        dE=dEpermm
                        F=2700.000

                    X=round(xc+pathRadius*math.cos(theta),3)
                    Y=round(yc+pathRadius*math.sin(theta),3)
                    E=round(E+dE*(pathRadius*2*math.pi/stepsInCircle),3)
                    sendCommand(C,X,Y,Z,F,E)


def block(pathAngleDeg,x0,y0,z0,xLength,yLength,height):
    global E


    if xLength>yLength:
        print("To save me a lot of coding, just set xLength<yLength.")
        return
    
    pathAngleDeg=float(pathAngleDeg)
    pathAngleRad=pathAngleDeg*2.0*math.pi/360.0

    sliceCount=math.ceil(height/layerHeight)
    C='G1' #'Controlled Move' extrude-move G-code
    F=2700.000

    #Retract extruder:
    gcode.write('G1 F9000.000 E-2.500\n')
    gcode.write('G1 F9000.000 E0.0\n')
    if pathAngleDeg<0.0:
        print("pathAngleDeg must be >=0 and <=90")

    if pathAngleDeg>90.0:
        print("pathAngleDeg must be >=0 and <=90")

    if pathAngleDeg==0.0:
        xSlicePointCount=math.ceil((xLength/fibreWidth))
        ySlicePointCount=0
        xHalfSlice=xSlicePointCount/2.0
        for sliceIterate in np.arange(1,sliceCount+1):
            Z=z0+sliceIterate*layerHeight
            gcode.write('G92 E0\n')
            E=0.0
            for passNumber in np.arange(xHalfSlice):
                xRel=fibreWidth*passNumber*2.0
                #One 'pass' is two roads width: (1)start, (2) extrude to the end, (3) move to the next roads end, (4) extrude back to start.
                #First point of pass:
                X=x0+xRel
                Y=y0
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                #Second point of pass:
                Y=y0+yLength
                E=round(E+dEpermm*yLength,5)
                sendCommand(C,X,Y,Z,F,E)
                #Third point of pass:
                E=round(E+dEpermm*fibreWidth*2.0,5)
                X=X+fibreWidth
                sendCommand(C,X,Y,Z,F,E)
                #Fourth point of pass:
                Y=y0
                E=round(E+dEpermm*yLength,5)
                sendCommand(C,X,Y,Z,F,E)
            
    if pathAngleDeg==90.0:
        xSlicePointCount=0
        ySlicePointCount=math.ceil((yLength/fibreWidth))
        yHalfSlice=ySlicePointCount/2.0
        for sliceIterate in np.arange(1,sliceCount+1):
            Z=z0+sliceIterate*layerHeight
            gcode.write('G92 E0\n')
            E=0.0
            for passNumber in np.arange(yHalfSlice):
                yRel=fibreWidth*passNumber*2.0
                #One 'pass' is two roads width: (1)start, (2) extrude to the end, (3) move to the next roads end, (4) extrude to next start position.
                #First point of pass:
                Y=y0+yRel
                X=x0
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                #Second point of pass:
                X=x0+xLength
                E=round(E+dEpermm*xLength,5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Third point of pass:
                E=round(E+dEpermm*fibreWidth*2.0,5)
                Y=Y+fibreWidth
                sendCommand(C,X,Y,Z,F,E)
                
                #Fourth point of pass:
                X=x0
                E=round(E+dEpermm*xLength,5)
                sendCommand(C,X,Y,Z,F,E)

    if pathAngleDeg>0.0 and pathAngleDeg<90.0:
        xSlicePointCount=math.floor((xLength/fibreWidth)*math.cos(pathAngleRad))
        ySlicePointCount=math.floor((yLength/fibreWidth)*math.sin(pathAngleRad))
        #Make xSlicePointCount and ySlicePointCount even:
        if xSlicePointCount%2==1:
            xSlicePointCount=xSlicePointCount+1
        if ySlicePointCount%2==1:
            ySlicePointCount=ySlicePointCount+1
            
        xHalfSlice=xSlicePointCount/2.0
        xBottomPoints=np.array([])
        xTopPoints=np.array([])
        yLeftPoints=np.array([])
        yRightPoints=np.array([])
        
        for i in np.arange(xSlicePointCount):
            xBottomPoints=np.append(xBottomPoints,[x0+i*xLength/xSlicePointCount])
            xTopPoints=np.append(xTopPoints,[x0+i*xLength/xSlicePointCount])
        
        for i in np.arange(ySlicePointCount):    
            yLeftPoints=np.append(yLeftPoints,[y0+yLength-i*yLength/ySlicePointCount])
            yRightPoints=np.append(yRightPoints,[y0+yLength-i*yLength/ySlicePointCount])

        for sliceIterate in np.arange(1,sliceCount+1):
            Z=z0+sliceIterate*layerHeight
            gcode.write('G92 E0\n')
            E=0.0

            #STAGE 1: Infill Top-Left Corner:
            X=x0
            Y=y0+yLength
            for passNumber in np.arange(xHalfSlice):
                #One 'pass' is two roads width: (1)start, (2) extrude to the end, (3) move to the next roads end, (4) extrude to next start position.
                #First point of pass:
                countIn=passNumber*2
                X=x0
                Y=yLeftPoints[countIn]
                xOld=X
                yOld=Y
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Second point of pass:
                X=xTopPoints[countIn]
                Y=y0+yLength
                E=round(E+dEpermm*(math.sqrt((X-xOld)**2.0+(Y-yOld)**2.0)),5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Third point of pass:
                X=xTopPoints[countIn+1]
                xOld=X
                yOld=Y
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Fourth point of pass:
                X=x0
                Y=yLeftPoints[countIn+1]
                E=round(E+dEpermm*(math.sqrt((X-xOld)**2.0+(Y-yOld)**2.0)),5)
                sendCommand(C,X,Y,Z,F,E)

            #STAGE 2: Infill middle (inherits X,Y coordinates from previous block):
            for passNumber in np.arange((ySlicePointCount-xSlicePointCount)/2):
                #One 'pass' is two roads width: (1)start, (2) extrude to the end, (3) move to the next roads end, (4) extrude back to start.
                #First point of pass:
                countIn=passNumber*2
                X=x0
                Y=yLeftPoints[countIn+xSlicePointCount]
                xOld=X
                yOld=Y
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Second point of pass:
                Y=yRightPoints[countIn]
                X=x0+xLength
                E=round(E+dEpermm*(math.sqrt((X-xOld)**2.0+(Y-yOld)**2.0)),5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Third point of pass:
                Y=yRightPoints[countIn+1]
                xOld=X
                yOld=Y
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Fourth point of pass:
                X=x0
                Y=yLeftPoints[countIn+1+xSlicePointCount]
                E=round(E+dEpermm*(math.sqrt((X-xOld)**2.0+(Y-yOld)**2.0)),5)
                sendCommand(C,X,Y,Z,F,E)

            yCountUsed=countIn+2
            #STAGE 3: Infill Bottom Right (inherits X,Y coordinates from previous block):
            for passNumber in np.arange(xHalfSlice):
                #One 'pass' is two roads width: (1)start, (2) extrude to the end, (3) move to the next roads end, (4) extrude back to start.
                #First point of pass:
                countIn=passNumber*2
                Y=y0
                X=xBottomPoints[countIn]
                xOld=X
                yOld=Y
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Second point of pass:
                X=x0+xLength
                Y=yRightPoints[countIn+yCountUsed]
                E=round(E+dEpermm*(math.sqrt((X-xOld)**2.0+(Y-yOld)**2.0)),5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Third point of pass:
                Y=yRightPoints[countIn+yCountUsed+1]
                xOld=X
                yOld=Y
                E=round(E+dEpermm*fibreWidth*2.0,5)
                sendCommand(C,X,Y,Z,F,E)
                
                #Fourth point of pass:
                X=xBottomPoints[countIn+1]
                Y=y0
                E=round(E+dEpermm*(math.sqrt((X-xOld)**2.0+(Y-yOld)**2.0)),5)
                sendCommand(C,X,Y,Z,F,E)

    #Raise Z-axis:
    gcode.write('G91\n')
    gcode.write('G1 Z50 F50\n')
    gcode.write('G90\n')


def endGcode():
    #Retract code:
    global E
    E=E-2.500
    gcode.write('G1 F9000.000 E')
    gcode.write(str(E))
    gcode.write('\n')

    #Set E datum to 0:
    gcode.write('G92 E0\n')

    #Raise Z-axis:
    gcode.write('G91\n')
    gcode.write('G1 Z50 F50\n')
    gcode.write('G90\n')

    #Home All Axes
    gcode.write('G28 X0 Y0\n')
    
    gcode.close()
###################### END OF OGCODE DEFINITIONS ######################
    
######################### Specimen Definition #########################
#Cylinder parameters: pathing,xc,yc,zc,innerRadius,outerRadius,height
#Block parameters: pathAngleDeg,x0,y0,z0,xLength,yLength,height
    
startGcode()

#Small Printer Calibration Test Block:
#block(0.0,100.0,50.0,0.0,10.0,10.0,2.0)

#Straight Cylindrical Pressure Vessel using two cylinders:
#Will produce cylinder with 10mm thick cap-end and 10mm ID.
#cylinder('concentric',100.0,100.0,0.0,0.0,10.0,10.0)
#cylinder('concentric',100.0,100.0,10.0,5.0,10.0,20.0)

#0 degree ASTMD3039 Specimen in middle of build area:
#block(0.0,90.0,7.5,0.0,25.0,185.0,2.0)

#10 degree ASTMD3039 Specimen in middle of build area:
#block(10.0,90.0,7.5,0.0,25.0,185.0,2.0)

#45 degree ASTMD3039 Specimen in middle of build area:
#block(45.0,90.0,7.5,0.0,25.0,185.0,2.0)

#60 degree ASTMD3039 Specimen in middle of build area:
block(60.0,90.0,7.5,0.0,25.0,185.0,2.0)

#90 degree ASTMD3039 Specimen in middle of build area:
#block(90.0,90.0,7.5,0.0,25.0,185.0,2.0)

endGcode()


